﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Threading;
using System.Drawing.Imaging;
using System.IO;
using AForge.Video.DirectShow;
using System.Net;

namespace VP8.NET.Example
{
    public partial class VideoChat : Form
    {
        private delegate void UIAction();

        private const int VideoWidth = 320;
        private const int VideoHeight = 240;

        private readonly EncodingSession EncodingSession = new EncodingSession(VideoWidth, VideoHeight, 15);
        private readonly DecodingSession DecodingSession = new DecodingSession();

        public VideoChat()
        {
            InitializeComponent();

            var deviceSelection = new DeviceSelection();
            deviceSelection.DeviceSelected += DeviceSelected;
            deviceSelection.ShowDialog();
        }

        private void DeviceSelected(object sender, DeviceSelectionEventArgs e)
        {
            // start capture
            var capture = new Capture(e.VideoDevice, VideoWidth, VideoHeight);
            capture.ImageCapture += new EventHandler<CaptureEventArgs>(ImageCapture);
            capture.Start();

            // when form closes, stop capture
            FormClosing += (s, ev) =>
            {
                capture.Stop();
            };
        }

        private void ImageCapture(object sender, CaptureEventArgs e)
        {
            // display image
            UpdateImage(picA, e.Image);

            // encode frame
            var data = EncodingSession.Encode(e.Image);
            if (data != null)
            {
                var image = DecodingSession.Decode(data);
                UpdateImage(picB, image);
            }
        }

        private void UpdateImage(PictureBox box, Image image)
        {
            try
            {
                using (var g = box.CreateGraphics())
                {
                    g.DrawImage(image, new Point(0, 0));
                }
            }
            catch (Exception) { }
        }
    }
}
